
if {$tcl_patchLevel ne "8.5.9"} {error "requires 8.5.9"}
if {![info exists ::vfs::tclkit_version]} {error "requires tclkit"}

# ==title== program

package require Tk

set dir [file dirname [info script]]
lappend auto_path [file join $dir lib]
set sourcefn [info script]

if {"console" ni [info commands]} {
  proc console {args} {} ;# do nothing
}

proc p {x} {xco_p $x}



####xia####platform####

proc windows {body} {
  if {$::tcl_platform(os) eq "Windows NT"} {
    uplevel $body
  }
}

proc linux {body} {
  if {$::tcl_platform(os) eq "Linux"} {
    uplevel $body
  }
}

proc osx {body} {
  if {$::tcl_platform(os) eq "Darwin"} {
    uplevel $body
  }
}


linux {


# metacity fix so that [raise] works properly
# http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=514115
#
#
rename toplevel Real_toplevel
proc toplevel {args} {
  uplevel [list eval Real_toplevel $args]
  wm group [lindex $args 0] .
}


# Text pasting doesn't overwrite the currently selected text,
# which is not what im used to.
proc proper_paste {w} {
  if {[catch {::tk::GetSelection $w CLIPBOARD} sel]} return
  catch {$w delete sel.first sel.last}
  $w insert insert $sel
}
bind Text <<Paste>> {proper_paste %W ; break}


} ;# end linux


proc maximize_zoom {w} {
  linux {wm attr . -zoomed 1}
  windows {wm state . zoomed}
}


####ixa####



####xia####xco####

namespace eval ::xco:: {
  variable ::xco::e_result {}
  variable ::xco::e_options {}
}

proc xco_vars {} {
  foreach v [info vars ::xco::*] {
    uplevel [list upvar $v [namespace tail $v]]
  }
}

proc xco_init {} {

  namespace eval ::xco:: {
    variable t_in .in_xco
    variable t_out .out_xco
    variable in ${t_in}.tx
    variable out ${t_out}.tx
  }

  xco_vars


  #
  # build 'in'
  #

  toplevel $t_in
  wm protocol $t_in WM_DELETE_WINDOW xco_hide

  pack [frame $t_in.ftop -bg purple -height 4] -side top -fill x
  text $in
  pack $in -fill both -expand 1

  xco_setup_tx $in
  wm geom $t_in 500x300
  focus $in

  bind $in <Control-Key-Return> [list xco_eval]
  bind $in <Control-KP_Enter>   [list xco_eval]

  bind $in <Control-w> {xco_hide ; break}


  #
  # build 'out'
  #

  toplevel $t_out
  wm protocol $t_out WM_DELETE_WINDOW xco_hide

  pack [frame $t_out.ftop -bg orange -height 4] -side top -fill x

  scrollbar $t_out.sb -orient vertical -command [list $out yview]
  pack $t_out.sb -fill y -side right

  text $out -yscrollcommand [list $t_out.sb set]
  pack $out -fill both -expand 1

  xco_setup_tx $out
  wm geom $t_out 600x400
  focus $out

  $out tag conf tag_timestamp -foreground cyan
  $out tag conf tag_ok -foreground yellow
  $out tag conf tag_error -foreground magenta

  xco_set_read_only $out


  bind $out <Control-w> {xco_hide ; break}


  # at the end

  wm geom $t_in +82+121
  wm geom $t_out +416+47

  xco_hide

  after idle [list xco_add_possible_bindings $in]
  after idle [list xco_add_possible_bindings $out]

}

proc xco_show {} {
  xco_vars ; wm deiconify $t_out ; wm deiconify $t_in
  raise $t_out ; raise $t_in ; focus -force $in
}
proc xco_hide {} {
  xco_vars ; wm withdraw $t_in ; wm withdraw $t_out
}

proc xco_thread_name {name} {
  xco_vars
  wm title $t_in "$t_in $name"
  wm title $t_out "$t_out $name"
}

proc xco_eval {} {
  xco_vars
  set c [$in get 1.0 end-1c]
  xco_eval_aux $c
  return -code break
}

proc xco_eval_aux {textcode} {
  xco_vars

  if {[catch [list uplevel #0 $textcode] ::xco::e_result ::xco::e_options]} {
    set tag tag_error
  } else {
    set tag tag_ok
  }

  $out ins end "\[[now]\] " tag_timestamp $::xco::e_result $tag \n
  $out see end

}

proc xco_p {x} {
  xco_vars
  $out ins end $x\n
  $out see end
}

proc xco_get_font {} {
  set priority_fonts {"Lucida Console" "Liberation Mono" "DejaVu Sans Mono"}
  set ffamilies [font families]
  foreach el $priority_fonts {
    if {$el in $ffamilies} {
      return $el
    }
  }
  return ""
}

proc xco_setup_tx {tx} {

  set pfont [xco_get_font]
  if {$pfont ne ""} {$tx conf -font [list $pfont 14]}

  $tx conf -undo 1 -wrap word \
    -bg #444444 -fg white -bd 0 \
    -insertwidth 4 -insertbackground green \
    -insertofftime 200 -highlightthickness 0

  # control-shift-a  is select all
  bind $tx <Control-A> {%W tag add sel 1.0 end-1c}

  # control-shift-z is redo, and we already have control-z for undo
  bind $tx <Control-Z> {event generate %W <<Redo>>}

  # two spaces instead of tabs
  bind $tx <Tab> {%W insert "insert" "  " ; break}

  # font resizing
  bind $tx <F3> [list xco_decr_font $tx]
  bind $tx <F3> +break
  bind $tx <F4> [list xco_incr_font $tx]
  bind $tx <F4> +break

}

proc xco_add_possible_bindings {tx} {

  if {[xco_proc_exists tx_scroll_setup_bind]} {
    tx_scroll_setup_bind $tx
  }

  if {[xco_proc_exists tx_center_line]} {
    bind $tx <Control-l> [list tx_center_line $tx]
  }

  if {[xco_proc_exists ::txcursor::group_forward]} {
    bind $tx <Alt-b> [list ::txcursor::group_backward $tx]
    bind $tx <Alt-f> [list ::txcursor::group_forward  $tx]
  }

  if {[xco_proc_exists tx_jump_to_begin]} {
    # emacs like alt-> alt-<
    bind $tx <Alt-KeyPress-less> [list tx_jump_to_begin $tx]
    bind $tx <Alt-KeyPress-less> +break
    bind $tx <Alt-KeyPress-greater> [list tx_jump_to_end $tx]
    bind $tx <Alt-KeyPress-greater> +break
  }

}

proc xco_proc_exists {x} {
  if {[catch {info body $x}]} {return 0} else {return 1}
}


proc xco_decr_font {w} {
  set font [$w cget -font]
  set size [lindex $font 1]
  if {$size == 1} return
  incr size -1
  $w conf -font [lreplace $font 1 1 $size]
}
proc xco_incr_font {w} {
  set font [$w cget -font]
  set size [lindex $font 1]
  incr size
  $w conf -font [lreplace $font 1 1 $size]
}

proc xco_set_read_only {tx} {
  rename $tx REAL_ONE_$tx
  proc $tx {args} {
    set tx [lindex [info level 0] 0]  ;# this proc name
    set four [string range $args 0 3]
    if {($four eq "inse") || ($four eq "dele")} return
    uplevel [list eval REAL_ONE_$tx $args]
  }
  $tx conf -insertwidth 0
}

####ixa####


####xia####winconsole####
console eval {
  bind Console <Control-Key-p> {tk::ConsoleHistory prev}
  bind Console <Control-Key-n> {tk::ConsoleHistory next}
  .console conf -font {"Lucida Console" 14} -bg #dddddd
  destroy .menubar
}
####ixa####

####xia####alt-menu-disable####
proc alt_menu_disable {} {
# remove global bindings that do the alt- menu (specific to this interp)
foreach el [bind all] {
  if {[bind all $el] eq "tk::WinMenuKey %W %N"} {
    bind all $el {}
  }
}
}
alt_menu_disable
####ixa####

####xia####verybase####
proc timestamp_format {x} {clock format $x -format "%Y-%m-%d--%H-%M-%S"}
proc now {} {timestamp_format [clock sec]}
proc sconcat {args} {append x {*}$args}
proc file_get {fn} {set f [open $fn r] ; set c [read $f] ; close $f ; return $c}
# save text files with lf encoding (unix encoding)
proc file_save {fn x} {set f [open $fn w] ; fconfigure $f -translation lf ; puts -nonewline $f $x ; close $f}
proc bin_file_get {fn} {set f [open $fn r] ; fconfigure $f -translation binary -encoding binary ; set c [read $f] ; close $f ; return $c}
proc bin_file_save {fn x} {set f [open $fn w] ; fconfigure $f -translation binary -encoding binary ; puts -nonewline $f $x ; close $f}
proc ? {} {set ::errorInfo}
####ixa####

####xia####verybase-bin####
proc to_hex {bytearray} {binary scan $bytearray H* hex; return $hex}
proc from_hex {hex} {set bytearray [binary format H* $hex]}
proc u32_to_bin {decimal} {set bytearray [binary format Iu $decimal]}
proc u32_to_dec {bytearray} {binary scan $bytearray Iu decimal ; return $decimal}

proc fbin_offsets {fn begin end} {
  set len [expr $end - $begin + 1]
  set f [open $fn r]
  fconfigure $f -translation binary -encoding binary
  seek $f $begin
  set data [read $f $len]
  close $f
  return $data
}
####ixa####

####xia####dict-utils####
proc dict_set_vars {d args} {
  foreach vname $args {
    set value [dict get $d $vname]
    uplevel [list set $vname $value]
  }
}

proc dget {d key elseval} {
  if {[dict exists $d $key]} {
    return [dict get $d $key]
  }
  return $elseval
}
####ixa####

####xia####eq-alias####
interp alias {} eq {} string equal
####ixa####

####xia####nsprocs####
proc link {args} {
  set vname [lindex $args 0]
  set ns [uplevel {set thisns}]
  set vns ${ns}::$vname
  uplevel [list upvar $vns $vname]
  if {[llength $args] == 2} {set $vns [lindex $args 1]}
}

proc links {args} {
  foreach vn $args {
    uplevel [list link $vn]
  }
}

proc linkall {} {
  set ns [uplevel {set thisns}]
  uplevel [list ns_upvar_all $ns]
}

# call from ::bim::some_proc, and it will link all namespace variables.
#
proc link_ns_vars {} {

  # ::bim
  set ns [uplevel {namespace current}]

  # ::bim::alpha ::bim::beta
  foreach fq_v [info vars ${ns}::*] {

    # alpha
    set v [namespace tail $fq_v]

    # uplevel {upvar ::bim::alpha alpha}
    uplevel [list upvar $fq_v $v]
  }

  return
}

proc nsvalue {ns varname} {
  set [sconcat $ns :: $varname]
}

proc ns_upvar {ns args} {
  foreach vname $args {
    set vns ${ns}::$vname
    uplevel [list upvar #0 $vns $vname]
  }
}

proc ns_upvar_all {ns} {

  # ::bim

  # ::bim::alpha ::bim::beta
  foreach fq_v [info vars ${ns}::*] {

    # alpha
    set v [namespace tail $fq_v]

    # uplevel {upvar ::bim::alpha alpha}
    uplevel [list upvar $fq_v $v]
  }

  return
}
####ixa####


####xia####lapin-rand####

# 0 to n-1 inclusive
proc lapin_rand_zto {n} {
  expr {int(rand() * $n)}
}

# a to b inclusive
proc lapin_rand_range {a b} {
  set c [lapin_rand_zto [expr {$b - $a + 1}]]
  expr {$a + $c}
}

# random element from list
proc lapin_rand_el {z} {
  set ix [lapin_rand_zto [llength $z]]
  lindex $z $ix
}

proc lapin_rand_color {} {
  format #%06x [lapin_rand_range 0 0xFFFFFF]
}
####ixa####


####xia####r9ine-font####

proc r9ine_decr_font {w} {

  set font [$w cget -font]
  set size [lindex $font 1]

  if {$size == 1} return

  incr size -1

  $w conf -font [lreplace $font 1 1 $size]

}

proc r9ine_incr_font {w} {

  set font [$w cget -font]
  set size [lindex $font 1]

  incr size

  $w conf -font [lreplace $font 1 1 $size]

}


proc r9ine_resize_tags {tx spec} {

  set tx_size [lindex [$tx cget -font] 1]

  foreach {tag diff} $spec {
    set tag_font [$tx tag cget $tag -font]
    set tag_new_size [expr {max(1, ($tx_size + $diff))}]
    $tx tag conf $tag -font [lreplace $tag_font 1 1 $tag_new_size]
  }

}


proc r9ine_tag_resize_spec {tx} {

  set tx_font [$tx cget -font]
  set tx_size [lindex $tx_font 1]

  set tag_resize_spec {}

  foreach tag [$tx tag names] {
    if {$tag eq "sel"} continue
    set tag_font [$tx tag cget $tag -font]
    set tag_size [lindex $tag_font 1]
    lappend tag_resize_spec $tag [expr {$tag_size - $tx_size}]
  }

  return $tag_resize_spec
}

####ixa####


####xia####txcursor####
namespace eval ::txcursor:: {}

proc ::txcursor::jump_cursor {tx ix} {
  $tx mark set insert $ix
  $tx see insert
}

proc ::txcursor::is_whitespace {s} {
  expr {[string is space -strict $s]}
}

proc ::txcursor::get_char_forward {tx ix} {
  $tx get $ix [sconcat $ix +1c]
}
proc ::txcursor::get_char_backward {tx ix} {
  $tx get [sconcat $ix -1c] $ix
}

proc ::txcursor::group_forward {tx} {

  set ix [$tx index insert]

  set s    [::txcursor::get_char_forward $tx $ix]
  set g_ws [::txcursor::is_whitespace $s]

  while {$ix ne "end"} {
    set s    [::txcursor::get_char_forward $tx $ix]
    set c_ws [::txcursor::is_whitespace $s]

    if {$c_ws != $g_ws} break
    set ix [$tx index [sconcat $ix +1c]]
  }

  ::txcursor::jump_cursor $tx $ix
}


proc ::txcursor::group_backward {tx} {

  set ix [$tx index insert]

  set s    [::txcursor::get_char_backward $tx $ix]
  set g_ws [::txcursor::is_whitespace $s]

  while {$ix ne "1.0"} {
    set s    [::txcursor::get_char_backward $tx $ix]
    set c_ws [::txcursor::is_whitespace $s]

    if {$c_ws != $g_ws} break
    set ix [$tx index [sconcat $ix -1c]]
  }

  ::txcursor::jump_cursor $tx $ix
}
####ixa####



####xia####tx_utils####

proc tx_num_matches {tx findtext} {
  set c [$tx get 1.0 end]
  llength [string_find_matches $c $findtext]
}

proc tx_center_line {tx} {
  lassign [$tx bbox insert] _ iy _ _
  set TY [winfo height $tx]
  set distance_to_middle [expr $iy - ($TY / 2.0)]
  $tx yview scroll $distance_to_middle pixels
}

proc tx_jump_to_pos {tx pos} {
  $tx mark set insert $pos
  $tx see insert
}
proc tx_jump_to_begin {tx} {
  tx_jump_to_pos $tx 1.0
}
proc tx_jump_to_end {tx} {
  tx_jump_to_pos $tx end
}

proc tx_jump_to_random_line {tx} {
  set line [$tx index end]
  set last [expr int(floor($line))]
  set ix [sconcat [lapin_rand_range 1 $last] ".0"]
  tx_jump_to_pos $tx $ix
}

####ixa####



####xia####tx_scroll####

proc tx_scroll {tx n} {
  $tx yview scroll $n pixels
  tx_scroll__aux__if_insert_isnt_visible_set_to_center $tx
  event generate $tx <<TxScroll>>
}

proc tx_scroll__aux__if_insert_isnt_visible_set_to_center {tx} {
  lassign [$tx bbox insert] _ iy _ _
  set hei [winfo height $tx]
  if {($iy eq "") || ($iy < 0) || ([expr $hei - $iy] < 20)} {
    $tx mark set insert @0,[expr $hei / 2]
  }
}

proc tx_scroll_setup_bind {tx} {

# only for linux, buttons 4 and 5
bind $tx <4> {tx_scroll %W -60 ; break}
bind $tx <5> {tx_scroll %W 60 ; break}

bind $tx <MouseWheel> {
  if {%D >= 0} {
    tx_scroll %W -80
  } else {
    tx_scroll %W 80
  }
  break
}

bind $tx <Control-MouseWheel> {
  if {%D >= 0} {
    tx_scroll %W -160
  } else {
    tx_scroll %W 160
  }
  break
}

bind $tx <F1> [list tx_scroll $tx -40]
bind $tx <F2> [list tx_scroll $tx 40]
bind $tx <F11> [list tx_scroll $tx -80]
bind $tx <F12> [list tx_scroll $tx 80]
bind $tx <Control-F11> [list tx_scroll $tx -160]
bind $tx <Control-F12> [list tx_scroll $tx 160]

} ;# end tx_scroll_setup_bind

####ixa####

####xia####tx_modified####

proc tx_modified {tx code} {
  upvar ::tx_modified_after_$tx after_modified
  set after_modified {}
  bind $tx <<Modified>> [list tx_modified_callback $tx $code]
}

proc tx_modified_callback {tx code} {
  upvar ::tx_modified_after_$tx after_modified
  ##p "[now] [clock milli] tx_modified_callback"
  after cancel $after_modified

  # events get collated together until some time has elapsed,
  # and only then, does the $code get run.
  # 
  set nxtime 20

  set after_modified [after $nxtime $code]
  $tx edit modified 0
}

####ixa####

####xia####tx_customize####

proc tx_configure_as_i_like {tx} {

  $tx conf -relief flat -undo 1 -wrap word \
    -insertwidth 4 -font {"Georgia" 14}

  $tx conf -bg black -fg #cccccc -insertbackground magenta
  $tx conf -inactiveselectbackground magenta -selectbackground magenta -selectforeground white

  $tx conf -highlightthickness 0

  tx_setup_full_bindings $tx

}

proc tx_setup_full_bindings {tx} {

  # control-shift-a  is select all
  bind $tx <Control-A> {%W tag add sel 1.0 end-1c}

  # control-shift-z is redo, and we already have control-z for undo
  bind $tx <Control-Z> {event generate %W <<Redo>>}

  # two spaces instead of tabs
  bind $tx <Tab> {%W insert "insert" "  " ; break}

  # alt-b and alt-f jump words like emacs
  bind $tx <Alt-b> [list ::txcursor::group_backward $tx]
  bind $tx <Alt-f> [list ::txcursor::group_forward  $tx]

  # center vertically
  bind $tx <Control-l> [list tx_center_line $tx]

  # emacs like alt-> alt-<
  bind $tx <Alt-KeyPress-less> [list tx_jump_to_begin $tx]
  bind $tx <Alt-KeyPress-less> +break
  bind $tx <Alt-KeyPress-greater> [list tx_jump_to_end $tx]
  bind $tx <Alt-KeyPress-greater> +break

  # use old style font specs, they bloody work correctly and fast
  bind $tx <F3> [list r9ine_decr_font $tx]
  bind $tx <F4> [list r9ine_incr_font $tx]

  # tx_scroll
  tx_scroll_setup_bind $tx

}

####ixa####


####xia####EEVAL####
proc EEVAL {} {
  toplevel .eeval
  set tx .eeval.tx

  text $tx -bg #eeeeee -fg black -font {{Lucida Console} 14} \
    -insertbackground black -insertwidth 4 -wrap none -undo 1
  pack $tx -fill both -expand 1

  wm geom .eeval 785x400

  focus -force $tx

  # control-shift-a  is select all
  bind $tx <Control-A> {%W tag add sel 1.0 end-1c}

  # control-shift-z is redo, and we already have control-z for undo
  bind $tx <Control-Z> {event generate %W <<Redo>>}

  # two spaces instead of tabs
  bind $tx <Tab> {%W insert "insert" "  " ; break}

  bind $tx <Control-Return> EEVAL_do
  bind $tx <Control-Return> +break

}
proc EEVAL_do {} {
  set tx .eeval.tx

  set c \n
  append c [string trim [$tx get 1.0 end-1c]]
  append c \n

  console eval [list tk::ConsoleInsert .console $c]
  console eval {
    tk::ConsoleInvoke
    .console mark set insert end-1c
  }
}

bind all <Alt-F12> EEVAL
####ixa####


####xia####string-utils####

proc string_num_matches {c findtext} {
  llength [string_find_matches $c $findtext]
}

proc string_find_matches {c findtext} {
  set z {}
  set si 0
  while 1 {
    set ix [string first $findtext $c $si]
    if {$ix == -1} break
    lappend z $ix
    set si [expr $ix + [string length $findtext]]
  }
  return $z
}

# x is given, replace first occurence of string a with string b.
proc string_replace_one {x a b} {
  set ix [string first $a $x]
  if {$ix == -1} {return $x}
  set ix_end [expr $ix + [string length $a] - 1]
  string replace $x $ix $ix_end $b
}

# x is given, replace all instances of a with b.
proc string_replace_all {x a b} {
  string map [list $a $b] $x
}

proc string_split_on_matches {c findtext} {
  set z {}
  set si 0
  while 1 {
    set ix [string first $findtext $c $si]
    if {$ix == -1} break
    lappend z [string range $c $si [expr $ix - 1]]
    set si [expr $ix + [string length $findtext]]
  }
  lappend z [string range $c $si end]
  return $z
}

proc string_from_this_until_end {c s0} {
  set i0 [string first $s0 $c]
  if {$i0 == -1} {return {}}
  set ni [expr $i0 + [string len $s0]]
  string range $c $ni end  
}

proc string_between {c s0 s1} {
  set i0 [string first $s0 $c]
  if {$i0 == -1} {return {}}
  set ni [expr $i0 + [string len $s0]]
  set i1 [string first $s1 $c $ni]
  if {$i1 == -1} {return {}}
  string range $c $ni [expr $i1 - 1]
}

proc all_strings_between {c xbegin xend} {
  set z {}
  set si 0
  while 1 {
    set ix [string first $xbegin $c $si]
    if {$ix == -1} break
    set si [expr $ix + [string length $xbegin]]
    set b0 $si

    set ix [string first $xend $c $si]
    if {$ix == -1} break
    set si [expr $ix + [string length $xend]]
    set b1 [expr $ix - 1]
  
    lappend z [string range $c $b0 $b1]
  }
  return $z
}

####ixa####



####xia####tx_jfind####

proc tx_jfind_back {tx findtext} {
  if {$findtext eq ""} return

  set c [$tx get 1.0 insert]
  set clen [string length $c]

  set abs_ix [lindex [string_find_matches $c $findtext] end]
  if {$abs_ix eq ""} return
  set kk [expr $clen - $abs_ix]

  $tx mark set insert "[$tx index insert] - $kk c"
  $tx see insert
}

proc tx_jfind_fwd {tx findtext} {
  if {$findtext eq ""} return

  set c [$tx get "insert + 1c" end]
  set clen [string length $c]

  set abs_ix [lindex [string_find_matches $c $findtext] 0]
  if {$abs_ix eq ""} return
  set kk [expr $abs_ix + 1]

  $tx mark set insert "[$tx index insert] + $kk c"
  $tx see "insert + 5lines"
}

####ixa####



####xia####iel_icons####

# 0 to n-1
proc iel_randn {n} {expr {int(rand() * $n)}}

proc iel_randcolor {} {format #%06x [expr {int(rand() * 0xFFFFFF)}]}

proc iel_randlist {L} {lindex $L [iel_randn [llength $L]]}



proc iel_shape_ell {img fg x y} {
  $img put $fg -to $x $y [expr $x+16] [expr $y+4]
  $img put $fg -to $x $y [expr $x+4]  [expr $y+16]
}

proc iel_shape_sqr {img fg x y} {
  $img put $fg -to $x $y [expr $x+8] [expr $y+8]
}

proc iel_paint_bg {img bg} {
  $img put $bg -to 0 0 40 40
}


proc iel_gen_paint_spec {} {

  set num_drawn_shapes 10

  set z {}
  for {set i 0} {$i < $num_drawn_shapes} {incr i} {
    set shape [iel_randlist {ell sqr}]
    set x [iel_randn 40] ; set y [iel_randn 40]
    lappend z $shape $x $y
  }

  return $z
}

proc iel_gen_full_icon_spec {} {
  set paint_spec [iel_gen_paint_spec]
  set fg [iel_randcolor] ; set bg [iel_randcolor]
  return [list [list $fg $bg] $paint_spec]
}


proc iel_paint_icon_given_spec {img full_icon_spec} {

  # get fg, bg, paint_spec
  foreach {fg bg} [lindex $full_icon_spec 0] break
  set paint_spec [lindex $full_icon_spec 1]

  iel_paint_bg $img $bg

  foreach {shape x y} $paint_spec {
    set shape_cmd iel_shape_$shape
    $shape_cmd $img $fg $x $y
  }

  return
}


proc iel_paint_icon_new_spec {img} {
  set spec [iel_gen_full_icon_spec]
  iel_paint_icon_given_spec $img $spec
  return $spec
}



proc iel_paint_border {img bdcolor} {

  foreach {x1 y1 x2 y2} {

     0  0  40  1

    39  0  40 40

     0 39  40 40

     0 39   1  1

  } {
    $img put $bdcolor -to $x1 $y1 $x2 $y2
  }

  return
}




proc iel_demo {} {

  set t .iel_demo
  toplevel $t
  wm geom $t 180x150
  $t conf -bg black

  pack [label $t.xi -border 0 -bg gray]

  set img [image create photo -height 40 -width 40]
  $t.xi conf -image $img

  bind $t <space> [list iel_demo_change $img]

}

proc iel_demo_change {img} {

  iel_paint_icon_new_spec $img
  iel_paint_border $img white

}


####ixa####



####xia####scrollify####
namespace eval ::scrollify:: {}

proc scrollify {given_c given_m} {
  set thisns ::scrollify::$given_c  ;# ns in the form ::scrollify::.c
  namespace eval $thisns {}

  link c $given_c ; link m $given_m
  link barpx 6

  bind $c <Destroy> [list scrollify_destroyed $thisns]

  $c conf -highlightthickness 0 -bg black

  $c create rectangle 0 0 0 0 -fill #888888 -width 0 -tags "yyybg Yside Items"
  $c create rectangle 0 0 0 0 -fill yellow -width 0 -tags "aaa Yside Items"

  $c create rectangle 0 0 0 0 -fill #888888 -width 0 -tags "xxxbg Xside Items"
  $c create rectangle 0 0 0 0 -fill yellow -width 0 -tags "qqq Xside Items"

  # panning box
  ##$c bind Yside <Enter> [list scrollify_yenter $thisns]

  # managed window
  $c create window 0 0 -window $m -anchor nw -width 0 -height 0 -tags $m

  bind $c <Configure> [list scrollify_reconfed $thisns]

  link response_time 25
  link afterYYY {}
  link afterXXX {}
  $m conf -yscrollcommand [list scrollify_yscrolled $thisns]
  $m conf -xscrollcommand [list scrollify_xscrolled $thisns]

}

proc scrollify_destroyed {thisns} {
  namespace delete $thisns
}

proc scrollify_reconfed {thisns} {
  linkall

  set w [winfo width $c]
  set h [winfo height $c]

  set mw [expr {$w - $barpx}]
  set mh [expr {$h - $barpx}]
  $c itemconfigure $m -width $mw -height $mh

  $c coords yyybg [expr {$w - $barpx}] 0 $w $h
  $c coords xxxbg 0 [expr {$h - $barpx}] $w $h

}

proc scrollify_yenter {thisns} {
  linkall

  set t $c.t

  if {[winfo exists $t]} {destroy $t}
  toplevel $t -bg magenta
  focus -force $t

  bind $t <Leave> [list destroy $t]
  ##bind $t <Control-Down> [list p %K]

  wm attr $t -alpha 0.5
  wm overrideredirect $t 1

  set x [winfo rootx $c]
  set y [winfo rooty $c]

  set px [winfo pointerx .]
  set py [winfo pointery .]

  set ox [expr $px - 20]
  set oy [expr $py - 20]
  wm geom $t 40x40+$ox+$oy

if 0 {
  set w [winfo width $c]
  set h [winfo height $c]

  set ox [expr $x + $w - 30]
  set oy $y

  wm geom $t 30x$h+$ox+$oy
}


}

proc scrollify_yscrolled {thisns b0 b1} {
  linkall
  after cancel $afterYYY
  set afterYYY [after $response_time [list scrollify_yscrolled_aux $thisns $b0 $b1]]
}
proc scrollify_yscrolled_aux {thisns b0 b1} {

  if {![namespace exists $thisns]} return  ;# was deleted

  linkall

  # if all is shown, zero dimensions
  if {($b0 eq "0.0") && ($b1 eq "1.0")} {
    $c coords aaa 0 0 0 0
    return
  }

  set w [winfo width $c]
  set h [winfo height $c]

  # height of percent displayed indicator
  set ph [expr round(($b1 - $b0) * $h)]

  lassign [list [expr $w - $barpx] 0 $w $ph ] x1 y1 x2 y2

  # offset by where the indicator is
  set offset [expr round($b0 * $h)]
  incr y1 $offset
  incr y2 $offset
  
  $c coords aaa $x1 $y1 $x2 $y2

}

proc scrollify_xscrolled {thisns b0 b1} {
  linkall
  after cancel $afterXXX
  set afterXXX [after $response_time [list scrollify_xscrolled_aux $thisns $b0 $b1]]
}
proc scrollify_xscrolled_aux {thisns b0 b1} {

  if {![namespace exists $thisns]} return  ;# was deleted

  linkall

  # if all is shown, zero dimensions
  if {($b0 eq "0.0") && ($b1 eq "1.0")} {
    $c coords qqq 0 0 0 0
    return
  }

  set w [winfo width $c]
  set h [winfo height $c]

  # width of percent displayed indicator
  set pw [expr round(($b1 - $b0) * $w)]

  set x1 0
  set y1 [expr $h - $barpx]
  set x2 $pw
  set y2 $h

  # offset by where the indicator is
  set offset [expr round($b0 * $w)]
  incr x1 $offset
  incr x2 $offset

  $c coords qqq $x1 $y1 $x2 $y2

}
####ixa####



# ==title== prograM BeginS

# now blocks end and the program begins


# ==title== keep track of special keys state

set ::ControlDOWN 0
bind all <KeyPress-Control_L>   {set ::ControlDOWN 1}
bind all <KeyRelease-Control_L> {set ::ControlDOWN 0}

set ::SuperDOWN 0
bind all <KeyPress-Super_L>   {set ::SuperDOWN 1 ; super_down %X %Y}
bind all <KeyRelease-Super_L> {set ::SuperDOWN 0}

set ::AltDOWN 0
if 0 {
bind all <KeyPress-Alt_L>   {set ::AltDOWN 1 ; super_down %X %Y}
bind all <KeyRelease-Alt_L> {set ::AltDOWN 0}
}

bind all <Motion> [list super_extra_all_motion %X %Y]


#  reset the special keys otherwise they stick when
#  coming in from a virtual desktop or alt-tabbing
proc reset_specials {} {
  set ::ControlDOWN 0
  set ::SuperDOWN 0
  set ::AltDOWN 0
}

proc init_specials {universe} {
  bind $universe <Unmap> reset_specials
  bind $universe <Map> reset_specials
  bind $universe <FocusOut> reset_specials
  bind $universe <FocusIn> reset_specials
}





# ==title== texts in universe (longview)

namespace eval ::longview:: {}

proc allns {} {
  namespace children ::longview::
}

proc w2ns {w} {
  return ::longview::$w
}

proc allw {} {
  # all windows in stacking order
  global universe
  foreach w [winfo children $universe] {
    if {[winfo class $w] eq "MilkJack"} {lappend z $w}
  }
  return $z
}

proc del_obj {thisns} {
  linkall
  destroy $f
  namespace delete $thisns
}

proc new_one_based {universe master source_tx} {
  after 10 [list new_one_based_aux $universe $master $source_tx]
}

proc new_one_based_aux {universe master source_tx} {
  set si [$source_tx index insert]
  set thisns [make_new_tx $universe $master]
  link tx
  after 10 [list yeah_insert $tx $si]
  # there is another nasty update somewhere that is causing this weirdness
  # probably after the dimensions of the window frame get set
}

proc yeah_insert {tx si} {
  $tx mark set insert $si
  $tx see insert
  ##tx_center_line $tx
}

proc identify_yourself {thisns} {
  linkall
  p "identifying"
  p "$thisns"
  p "$tx"
}

set ::last_focused_tx {}

proc make_new_tx {given_universe given_master} {

  set ewindow .milkjack
  set ens ::longview::
  set eclass MilkJack

  set frame_color [lapin_rand_color]

  set dotw [winfo width .]
  set doth [winfo height .]

  # if first child
  if {[llength [allns]] == 0} {
    set kwidth [expr $dotw - 100]
    set kheight [expr $doth - 100]
    set kx 10
    set ky 30
  } else {
    set kwidth [lapin_rand_range 200 600]
    set kheight [lapin_rand_range 200 400]
    set kx [lapin_rand_range 0 100]
    set ky [lapin_rand_range 0 100]
  }


  set thisns [make_movable_frame $given_universe [list \
    ewindow $ewindow ens $ens eclass $eclass \
    frame_color $frame_color \
    kwidth $kwidth kheight $kheight \
    kx $kx ky $ky \
  ]]

  linkall

  link master $given_master

  link c $ffc.c
  link tx $c.tx

  canvas $c
  $master peer create $tx
  scrollify $c $tx
  pack $c -fill both -expand 1

  tx_configure_as_i_like $tx
  $tx conf -font $::codefont
  $tx conf -bg black -fg white  ;# boost contrast
  $tx conf -wrap [$master cget -wrap]

  sel_make $thisns

  focus -force $tx

  bind $tx <FocusIn> [list focus_in $thisns]
  bind $tx <FocusOut> [list focus_out $thisns]

  bind $tx <Button-3> [list move_click $thisns %X %Y]
  bind $tx <B3-Motion> [list move_motion $thisns %X %Y]

  bind $tx <Control-space> [list new_one_based $universe $master $tx]
  bind $tx <Control-space> +break

  bind $tx <Control-Alt-w> [list toggle_wordwrap]
  bind $tx <Control-Alt-w> +break

  bind $tx <Control-Alt-i> [list identify_yourself $thisns]
  bind $tx <Control-Alt-i> +break

  bind $tx <Control-s> [list tx_control_s $thisns]
  bind $tx <Control-s> +break
  bind $tx <Control-r> [list tx_control_r $thisns]
  bind $tx <Control-r> +break
  bind $tx <Control-g> [list tx_control_g $thisns]
  bind $tx <Control-g> +break

  bind $tx <Escape> [list tx_escape]
  bind $tx <Escape> +break


  bind $tx <F7> [list jump_to_current $tx]
  bind $tx <Control-F7> [list put_current_here $tx]



  bind $tx <<Selection>> [list sel_changed $thisns]
  bind $tx <<TxScroll>> +[list sel_scrolled $thisns]


  # these three standard run/save/backup hotkeys use the control key.

  bind $tx <Control-F1> [list run_file]
  bind $tx <Control-F1> +break

  bind $tx <Control-F2> [list save_to_file]
  bind $tx <Control-F2> +break

  bind $tx <Control-F9> [list make_backup_file]
  bind $tx <Control-F9> +break

  bind $tx <Control-w> [list bye_window $thisns]


  return $thisns
}

proc get_tx {thisns} {
  link tx
  return $tx
}

if 0 {
run it like this:
change {$tx conf -insertbackground red}
only takes effect for what is opened... quick try outs.
}
proc change {body} {
  foreach ns [allns] {
    set tx [get_tx $ns]
      eval $body
  }
}

proc focus_on_tx {thisns} {
  linkall
  focus -force $tx
}

proc focus_on_next_milk {} {
  set zall [allw]
  if {$zall eq ""} return
  set topnext [lindex $zall end]
  set ns [w2ns $topnext]
  focus_on_tx $ns
}

proc bye_window {thisns} {
  if {[llength [allns]] == 1} return
  del_obj $thisns
  focus_on_next_milk
}

proc close_all {} {
  foreach ns [allns] {
    bye_window $ns
  }
}

proc set_pos {thisns x y} {
  linkall
  $universe coords $thisns $x $y
}

proc set_dims {thisns w h} {
  linkall
  $universe itemconf $thisns -width $w -height $h
}

proc raise_it {thisns} {
  linkall
  raise $f
}

proc focus_in {thisns} {
  link tx
  set ::last_focused_tx $tx
  ##$tx conf -bg navy
}

proc focus_out {thisns} {
  link tx
  ##$tx conf -bg black
}

proc tx_escape {} {
  slots_toggle
}

proc tx_control_s {thisns} {
  global master_findtext
  if {$master_findtext eq ""} {
    if {![slots_are_visible]} {slots_place}
    focus_on_last_slot
    return
  }
  link tx
  tx_jfind_fwd $tx $master_findtext
  event generate $tx <<TxScroll>>
}

proc tx_control_r {thisns} {
  global master_findtext
  if {$master_findtext eq ""} {
    return
  }
  link tx
  tx_jfind_back $tx $master_findtext
  event generate $tx <<TxScroll>>
}

proc tx_control_g {thisns} {
  slot_cancel
}


# ==title== jump to CURRENT string

# works perfectly, taken from naiia, saved for posterity

proc jump_to_current {tx} {
  set findtext [sconcat # CURRENT]
  set c [$tx get 1.0 end]
  set ix [lindex [string_find_matches $c $findtext] 0]
  if {$ix eq ""} return
  $tx mark set insert "1.0 + $ix c"
  $tx see insert
  tx_center_line $tx
}
#CURRENT
proc del_string_if_exists {tx findtext} {
  set c [$tx get 1.0 end]
  set ix [lindex [string_find_matches $c $findtext] 0]
  if {$ix eq ""} return
  $tx del "1.0 + $ix c" "1.0 + $ix c + [string length $findtext] c"
}

proc put_current_here {tx} {
  set findtext [sconcat # CURRENT]
  del_string_if_exists $tx $findtext
  set savedpos [$tx index insert]
  $tx insert "insert" $findtext
  $tx mark set insert $savedpos
}


# ==title== slots !

namespace eval ::slots:: {}

set the_last_slot ::slots::slot9

proc slots_awake {universe master} {

  set zns {}
  for {set i 0} {$i < 10} {incr i} {
    lappend zns [make_slot $universe $master $i]
  }

  # could restore the slot texts here before doing the update count

  foreach ns $zns {
    slot_update_count $ns
  }

}


proc make_slot {given_universe given_master given_nn} {

  set thisns [sconcat ::slots:: slot$given_nn]
  namespace eval $thisns {}

  link nn $given_nn
  link universe $given_universe
  link master $given_master

  # has to be a child above the universe so it overlays always
  set pp [winfo toplevel $universe]
  if {$pp eq "."} {set pp ""}
  link w .slot$nn

  frame $w -bd 1 -bg orange
  # the special slot
  if {$thisns eq "::slots::slot9"} {
    $w conf -bg magenta
  }

  link inactivebg #222222 ; link activebg red

  text $w.tx 
  tx_configure_as_i_like $w.tx
  $w.tx conf -font {"Lucida Console" 14} \
    -height 1 -width 17 -bg $inactivebg -fg white -insertbackground green

  pack $w.tx -side left

  bind $w.tx <Control-a> [list $w.tx mark set insert 1.0]
  bind $w.tx <Control-e> [list $w.tx mark set insert end]

  bind $w.tx <Control-u> [list slot_kill_text $thisns]  

  bind $w.tx <Control-s> [list slot_control_s]
  bind $w.tx <Control-g> [list slot_control_g]

  bind $w.tx <Return> [list slot_return_key $thisns]
  bind $w.tx <Return> +break

  bind $w.tx <Control-Return> [list slot_control_return_key $thisns]
  bind $w.tx <Control-Return> +break

  bind $w.tx <FocusIn> [list set the_last_slot $thisns]

  bind $w.tx <Escape> slots_forget

  bind $w.tx <Control-p> [list move_to_slot [expr $nn - 1]]
  bind $w.tx <Control-p> +break
  bind $w.tx <Control-n> [list move_to_slot [expr $nn + 1]]
  bind $w.tx <Control-n> +break

  tx_modified $w.tx [list slot_modified $thisns]

  label $w.hits -font {"Lucida Console" 14} -width 5 -height 1 -bg black -fg yellow -relief flat
  pack $w.hits -side right -padx 1

  $w.hits conf -text 436$nn

  slot_place_one $nn
  
  return $thisns
}

proc slots_toggle {} {
  if {[slots_are_visible]} {
    slots_forget
  } else {
    slots_place
  }
}

proc slots_are_visible {} {
  if {[place info .slot0] eq ""} {return 0} else {return 1}
}

proc slots_place {} {
  for {set i 0} {$i < 10} {incr i} {
    slot_place_one $i
  }
}

proc slots_forget {} {
  place forget .slot0
  # all subsequent shall be forgotten as they are relatively placed
}

proc slot_place_one {nn} {
  set w .slot$nn
  if {$nn == 0} {
    # place the first slot relative to the container
    place $w -anchor ne -relx 1.0 -x -10 -rely 0.0 -y 10
  } else {
    # place each subsequent slot relative to the previous slot
    set prevw .slot[expr $nn - 1]
    # the after idle is necessary so that the previous slot
    # has its size and position set down before we do a relative
    # positioning of this new slot.
    after idle [list place_next_slot $prevw $w]
  }
}

proc place_next_slot {prevw w} {
  set yoffset [expr [winfo reqheight $prevw] + 8]
  place $w -in $prevw -anchor nw -x 0 -y $yoffset
}

proc move_to_slot {slotn} {
  if {$slotn < 0} return
  if {$slotn > 9} return
  set thisns ::slots::slot$slotn
  linkall
  focus -force $w.tx
}

proc focus_on_last_slot {} {
  global the_last_slot
  if {$the_last_slot eq ""} return
  set thisns $the_last_slot
  linkall
  focus -force $w.tx
}

proc all_slot_ns {} {
  lsort [namespace children ::slots::]
}

proc all_slot_w {} {
  set z {}
  foreach ns [all_slot_ns] {
    lappend z [get_slot_w $ns]
  }
  return [lsort $z]
}

proc get_slot_w {thisns} {link w ; return $w}

proc slot_set_inactive_color {thisns} {
  linkall
  $w.tx conf -bg $inactivebg
}

proc slot_set_all_inactive_color {} {
  foreach ns [all_slot_ns] {
    slot_set_inactive_color $ns
  }
}

proc is_slot_active {thisns} {
  linkall
  set bg [$w.tx cget -bg]
  if {$bg eq $activebg} {return 1} else {return 0}
}

proc get_active_slot {} {
  foreach ns [all_slot_ns] {
    if {[is_slot_active $ns]} {return $ns}
  }
  return {}
}

proc slot_focus_on_milk {} {
  set w $::last_focused_tx
  if {[winfo exists $w]} {
    focus -force $w
  } else {
    focus_on_next_milk
  }
}

proc slot_return_key {thisns} {
  linkall
  slot_set_all_inactive_color ; $w.tx conf -bg $activebg
  set findtext [$w.tx get 1.0 end-1c]
  master_slot_set $findtext
  slot_focus_on_milk
}

proc slot_control_return_key {thisns} {
  slot_set_all_inactive_color
  master_slot_clear
}

proc slot_special_set {c} {
  set thisns ::slots::slot9
  linkall
  $w.tx del 1.0 end
  $w.tx ins end $c
}

proc slot_kill_text {thisns} {
  linkall
  $w.tx del 1.0 end
  if {[is_slot_active $thisns]} {
    slot_set_all_inactive_color
    master_slot_clear
  }
  set original_bg [$w.tx cget -bg]
  $w.tx conf -bg yellow
  after 200 [list $w.tx conf -bg $original_bg]
}

proc slot_modified {thisns} {
  linkall
  if {[is_slot_active $thisns]} {
    set findtext [$w.tx get 1.0 end-1c]
    master_slot_set $findtext
  }
  slot_update_count $thisns
}

proc slot_rerun_was_modified {} {
  foreach ns [all_slot_ns] {
    slot_modified $ns
  }
}

proc slot_control_s {} {
  slot_focus_on_milk
}

proc slot_cancel {} {
  slot_set_all_inactive_color
  master_slot_clear
}

proc slot_control_g {} {
  slot_set_all_inactive_color
  master_slot_clear
  slot_focus_on_milk
}

proc slot_update_count {thisns} {
  linkall
  set findtext [$w.tx get 1.0 end-1c]
  if {$findtext eq ""} {
    set x 0
  } else {
    set x [tx_num_matches $master $findtext]
  }
  $w.hits conf -text $x
}



# ==title== the selection and the magenta box

proc sel_make {thisns} {
  linkall
  link selafter {}
  link selw $tx.selw

  set w $selw
  frame $w -bd 1 -bg white -width 40 -height 40
  label $w.i -bg magenta -fg white -font {"Lucida Console" 14 bold}
  pack $w.i -fill both -expand 1
  $w.i conf -text 9

}

proc sel_changed {thisns} {
  linkall

  set selrange [$tx tag ranges sel]
  if {$selrange eq ""} {
    set selafter [after 100 [list place forget $selw]]
    return
  }

  lassign $selrange s0 s1
  set c [$tx get $s0 $s1]

  slot_special_set $c

  lassign [$tx bbox $s0] x0 y0

  $selw.i conf -text [string_num_matches [master_text] $c]

  after cancel $selafter

  sel_place_xy $selw $x0 $y0

}

proc sel_place_xy {selw x0 y0} {
  place $selw -x [expr $x0 + 0] -y [expr $y0 - 40]
}

proc sel_scrolled {thisns} {
  linkall

  set selrange [$tx tag ranges sel]
  if {$selrange eq ""} return

  lassign $selrange s0 s1
  lassign [$tx bbox $s0] x0 y0

  sel_place_xy $selw $x0 $y0

}


# ==title== state of text windows in a file

# this is only really good for while im working on it and
# reloading itself, so i can save my windows and restore them.
# its not something for general use.
# i guess it could be. i call save_info and restore_info from the console.
# its a bit off, the windows are slightly scrolled differently
# because it centers on the insert cursor.

proc get_props {thisns} {
  linkall
  lappend z insert [$tx index insert]
  lappend z font [$tx cget -font]
  lappend z w [winfo width $f] h [winfo height $f]
  lassign [$universe coords $f] x y
  lappend z x $x y $y
  lappend z fcolor [$ffc cget -bg]
  lappend z yscroll [lindex [$tx yview] 0]
  return $z
}

proc set_props {thisns zprop} {
  linkall
  dict_set_vars $zprop insert font w h x y fcolor yscroll

  $tx conf -font $font
  set_dims $thisns $w $h
  set_pos $thisns $x $y
  $ffc conf -bg $fcolor

  $tx mark set insert $insert
  $tx see insert

  # a little out of sync on restore because only jumps to insert
  # still good though.

  ## this doesnt seem to work:
  ##$tx yview moveto $yscroll

}

proc info_file {} {
  global XFILE
  if {$XFILE eq ""} {error "no file loaded yet"}
  return [file join [file dirname $XFILE] milkjack-window-state]
}

proc save_info {} {
  foreach w [allw] {
    set ns [w2ns $w]
    lappend z [get_props $ns]
  }
  file_save [info_file] $z
}

proc restore_info {} {
  set z [file_get [info_file]]
  close_all
  set orig [lindex [allns] 0]  ;# there is one window after closing all

  set num [llength $z]
  if {$num == 0} return

  foreach zprop $z {
    set ns [make_new_tx $::universe $::master]
    set_props $ns $zprop
  }

  bye_window $orig  ;# close that one window

}


# ==title== right click moves or grows windows

proc and_move_it_by {thisns dx dy} {
  linkall
  lassign [$universe coords $thisns] x y
  set x [expr {$x + $dx}]
  set y [expr {$y + $dy}]
  $universe coords $thisns $x $y
}

proc and_grow_it_by {thisns dx dy} {
  linkall
  set w [$universe itemcget $thisns -width]
  set h [$universe itemcget $thisns -height]
  set w [expr {$w + $dx}]
  set h [expr {$h + $dy}]
  $universe itemconf $thisns -width $w -height $h
}

proc tx_inside {thisns} {
  linkall
  info exists tx
}

proc move_click {thisns X Y} {
  global QX QY
  set QX $X
  set QY $Y
  raise_it $thisns
  if {[tx_inside $thisns]} {focus_on_tx $thisns}
}
proc move_motion {thisns X Y} {
  global QX QY

  set dx [expr {$X - $QX}]
  set dy [expr {$Y - $QY}]
  set QX $X
  set QY $Y

  if {$::ControlDOWN == 0} {
    and_move_it_by $thisns $dx $dy
  } else {
    and_grow_it_by $thisns $dx $dy
  }

}

proc what_im_over {X Y} {
  set w [winfo containing $X $Y]
  if {$w eq ""} {return ""}

  # make sure that the toplevel has focus
  if {[winfo toplevel $w] ne [winfo toplevel [focus]]} {return ""}

  return [component_of_what $w]
}

# ==title== super key moves

set QQNS {}
proc super_down {X Y} {
  global QQNS
  set QQNS [what_im_over $X $Y]
  if {$QQNS eq ""} return
  move_click $QQNS $X $Y
}

proc super_extra_all_motion {X Y} {
  if {$::SuperDOWN == 0 && $::AltDOWN == 0} return
  global QQNS
  if {$QQNS eq ""} return
  move_motion $QQNS $X $Y
}


# ==title== toggle wrapping on all

proc set_wrapping {thisns r} {
  linkall
  $tx conf -wrap $r
}
proc toggle_wordwrap {} {
  global master
  set r [$master cget -wrap]
  if {$r eq "word"} {set r "none"} else {set r "word"}
  $master conf -wrap $r
  foreach ns [allns] {
    set_wrapping $ns $r
  }
}

# ==title== generic make a movable frame in the universe

proc make_movable_frame {given_universe dops} {

  dict_set_vars $dops \
    ewindow ens eclass \
    frame_color kwidth kheight kx ky

  # ewindow is the prefix for the frame that is a child of the universe canvas.
  # ens is the namespace that thisns will be a child of.
  # eclass is the class of the frame that is a child of the universe canvas.
  #        with this class you can figure out what type of movable window you have.

  set clockn [clock micro]
  set the_frame [sconcat $given_universe $ewindow $clockn]
  
  set thisns [sconcat $ens $the_frame]
  namespace eval $thisns {}

  link universe $given_universe
  link id $clockn
  link f $the_frame

  # this frame is the one that is managed by universe
  #
  frame $f -class $eclass -bg $frame_color

  # ffc is the frame you draw into
  #
  link ffc
  set ffc $f.ffc
  frame $ffc
  pack $ffc -fill both -expand 1 -padx 2 -pady 2

  $universe create window $kx $ky -window $f \
    -anchor nw -width $kwidth -height $kheight \
    -tags [list $f $thisns]

  register_frame $f $thisns

  return $thisns
}

# ==title== frame registry

set RR {}
proc register_frame {f thisns} {
  global RR
  dict set RR $f $thisns
}

# given a window, what ns is it a component of
# esoteric, to be sure, but quite useful
proc component_of_what {w} {

  global RR

  foreach f [dict keys $RR] {
    if {[string first $f $w] == 0} {
      return [dict get $RR $f]
    }
  }

  return {}
}


# ==title== navigation window

namespace eval ::ocean:: {}

proc make_nav {given_universe} {

  set ewindow .ocean
  set ens ::ocean::
  set eclass Ocean

  set frame_color gold

  set kwidth 300
  set kheight 100
  set kx 30
  set ky 10

  set thisns [make_movable_frame $given_universe [list \
    ewindow $ewindow ens $ens eclass $eclass \
    frame_color $frame_color \
    kwidth $kwidth kheight $kheight \
    kx $kx ky $ky \
  ]]

  linkall

  link wnum $ffc.num
  label $wnum \
    -font {"Lucida Console" 22 bold} \
    -fg magenta -bg black \
    -text n/a -width 5
  pack $wnum -side right -fill y

  link c $ffc.c
  link tx $c.tx

  canvas $c
  text $tx
  scrollify $c $tx
  pack $c -fill both -expand 1

  tx_configure_as_i_like $tx
  $tx conf -font $::codefont
  $tx conf -bg black -fg white  ;# boost contrast
  $tx conf -wrap none

  focus -force $tx

  bind $tx <Button-3> [list move_click $thisns %X %Y]
  bind $tx <B3-Motion> [list move_motion $thisns %X %Y]

  bind $tx <Control-Alt-i> [list identify_yourself $thisns]
  bind $tx <Control-Alt-i> +break

  tx_modified $tx [list nav_modified $thisns]

  return $thisns
}

proc nav_modified {thisns} {

  linkall

  set c [master_text]

  set findtext [$tx get 1.0 end-1c]

  set num [llength [string_find_matches $c $findtext]]

  $wnum conf -text $num
}


set nfo_text {This program is yours to play with in whatever way you can.

#1: Drop a text file into it.

Control-Spacebar makes a new window view

Control-W closes that window

Move a window by right-clicking and dragging.
On linux you can also use the windows-key instead of right-clicking.

Resize a window by holding the control key while doing the same as above.

Escape toggles the search slots

Save file is Control-F2

Word-wrap toggle is Control-Alt-W

Select All is Control-Shift-A

F3 and F4 Change the font size

Don't like the font choice? Edit the code in the ".x" file

Scroll with your mousewheel, or press control-mousewheel to scroll even faster

F1 and F2 and F11 and F12 scroll like the mousewheel.


If you have an interest then edit the ".x" File

all is stored there

nothing is hidden


--
XRO Solar
xro@tigerbliss.com

http://SupremeXeN.com

}

# ==title== info window

namespace eval ::nfo:: {}

proc make_nfo {given_universe} {

  set ewindow .nfo
  set ens ::nfo::
  set eclass Nfo

  set frame_color magenta

  set kwidth 700
  set kheight 400
  set kx 30
  set ky 10

  set thisns [make_movable_frame $given_universe [list \
    ewindow $ewindow ens $ens eclass $eclass \
    frame_color $frame_color \
    kwidth $kwidth kheight $kheight \
    kx $kx ky $ky \
  ]]

  linkall

  link c $ffc.c
  link tx $c.tx

  canvas $c
  text $tx
  scrollify $c $tx
  pack $c -fill both -expand 1

  tx_configure_as_i_like $tx

  set fam CODON
  if {$fam ni [font families]} {set fam serif}

  $tx conf -font [list $fam 16]
  $tx conf -bg black -fg white
  $tx conf -wrap word

  focus -force $tx

  bind $tx <Button-3> [list move_click $thisns %X %Y]
  bind $tx <B3-Motion> [list move_motion $thisns %X %Y]

  bind $tx <Control-Alt-i> [list identify_yourself $thisns]
  bind $tx <Control-Alt-i> +break

  bind $tx <Control-w> [list $universe delete $thisns]

  $tx ins end $::nfo_text
  $tx mark set insert 1.0

  return $thisns
}





# ==title== thread procs

# this is the master thread error control proc.
# when an error is reported in a child thread, it should
# be set to call this proc.
#
# this proc will call the bgerror with info about the child thread.
#
proc main_thread_error {thread_name thread_id errorInfo} {
  set ::errorInfo $errorInfo
  bgerror "thread error $thread_name ($thread_id)"
}

proc new_thread {} {

  set t [thread::create]


  # threadBOSS global
  #
  thread::send $t [list set ::threadBOSS [thread::id]]


  # Tk is required so the errorproc can send errors to the main thread
  thread::send $t {package require Tk ; wm withdraw .}


  # set the error proc for the thread, it will call
  # the main thread with details of the error.
  #
  thread::send $t {
    proc child_error_proc {thread_id errorInfo} {
      thread::send $::threadBOSS [list main_thread_error TTX $thread_id $errorInfo]
    }
    thread::errorproc child_error_proc
  }


  # load up the thread with the full source code
  #
  # very very cool
  #
  thread::send $t {set is_thread 1}
  thread::send $t [list source $::sourcefn]



  return $t  
}



# ==title== extra fonts for windows

####xia####fonts####

proc setup_fonts {} {

  package require ffidl

  # session-private for this process only
  ffidl::callout add_private_font_ffidl {pointer-utf8 uint32 pointer} int [ffidl::symbol gdi32.dll AddFontResourceExA]

  add_font_dir [file join $::dir fonts]
}

proc add_private_font {fn} {
  set fn [file nativename [file normalize $fn]]
  add_private_font_ffidl $fn 0x10 0  ;# 0x10 is FR_PRIVATE and last arg must be zero according to msdn
}

proc add_font_dir {d} {
  set z [glob -nocomp -dir $d *.ttf]
  foreach el $z {add_private_font $el}
}

####xia####



# ==title== tkdnd drag and drop for windows

proc setup_dnd {dropw} {
  package require tkdnd
  set dtype CF_HDROP
  dnd bindtarget $dropw $dtype <DragEnter>  {return [lindex %A 0]}
  dnd bindtarget $dropw $dtype <Drag>       {return [lindex %A 0]}
  dnd bindtarget $dropw $dtype <Drop>       [list tkdnd_dropped $dropw %D %T %C]
  dnd bindtarget $dropw $dtype <DragLeave>  {}
}

proc unsetup_dnd {dropw} {
  set dtype CF_HDROP
  dnd bindtarget $dropw $dtype <DragEnter>  {}
  dnd bindtarget $dropw $dtype <Drag>       {}
  dnd bindtarget $dropw $dtype <Drop>       {}
  dnd bindtarget $dropw $dtype <DragLeave>  {}
}

proc tkdnd_dropped {dropw D T C} {
  # D T C => data type code
  unsetup_dnd $dropw
  set fn [lindex $D 0] ;# only look at the first item
  do_the_loadfile $fn
}


# ==title== savinggggg and runninggggg

proc show_saved {} {
  global universe
  set f $universe.saved
  label $f -text saved -font {Georgia 48 bold} -fg magenta -bg black
  update idletasks
  set w [winfo reqwidth $f]
  set h [winfo reqheight $f]
  $universe create window 10 10 -window $f -anchor nw -width $w -height $h -tags saved
  after 2000 [list destroy $f]
}

proc save_to_file {} {
  global XFILE
  file_save $XFILE [text_of_save]
  show_saved
}

proc text_of_save {} {
  global master
  set c [$master get 1.0 end-1c]
}

proc run_file {} {
  global XFILE
  set c [text_of_save]
  file_save $XFILE $c
  set ext ".exe"
  linux {set ext ""}
  set ename [sconcat [file rootname [file tail $XFILE]] $ext]
  set e [file join [file dirname $XFILE] $ename]
  if {[file exists $e]} {
    exec $e &
  }
}

proc do_the_backup {f} {
  global XFILE
  set txt [$f get 1.0 end-1c]
  set tm [now]
  set fn [sconcat $XFILE . $tm . $txt]
  file_save $fn [text_of_save]
  after 1000 [list destroy $f]
  focus_on_next_milk
}

proc make_backup_file {} {
  global universe
  set f $universe.bak
  text $f -font {"Lucida Console" 18} -bg black -fg red -width 22 -height 1 -insertbackground yellow
  update idletasks
  set w [winfo reqwidth $f]
  set h [winfo reqheight $f]
  $universe create window 10 10 -window $f -anchor nw -width $w -height $h -tags bak
  focus $f
  bind $f <Control-g> [list destroy $f]
  bind $f <Return> [list do_the_backup $f]
  bind $f <Return> +break
}

# ==title== master code

proc master_text {} {
  global master
  $master get 1.0 end-1c
}

proc len {} {
  global master
  string length [$master get 1.0 end-1c]
}

proc create_master_tags {} {
  global master
  # specifying font sizes upsets the font resizing of F3 and F4
  $master tag configure tag_title -foreground black -background magenta ;#-font {Georgia 24 bold}
  $master tag configure tag_xia -foreground black -background green ;#-font {Georgia 20 bold}
  $master tag configure tag_proc -foreground cyan
  $master tag configure tag_proc_right -foreground yellow ;#-font {"Lucida Console" 20}
  $master tag configure tag_search -foreground black -background cyan
}

set ::master_findtext {}

proc master_slot_set {findtext} {
  global master  ;# yes, global
  # why not global? am i even ever going to have
  # more 'universes' no. there is only one canvas.
  # at least there will only ever be for this program.
  # eroded by the winds, a definite form begins to take shape.
  # there is no going back.

  set ::master_findtext $findtext
  generic_autotag_string $master $findtext tag_search

}

proc master_slot_clear {} {
  global master
  set ::master_findtext {}
  $master tag remove tag_search 1.0 end
}


# ==title== parsing thread including autotagging

if 0 {

How does this work?

The TTX thread is created and seeded with some procs and other stuff to 
make it work. Once that happens it's ready to be used. Procs are called in
it that are always executed from the last one first. That's done by using
after events and cancelling the previous one. So in effect when a 
long running proc is doing its thing, any other calls to the start proc
will get queued with [after 10 proc] and then only the last one will 
execute when the long running proc is done and the interp thread is ready
to run other commands.

Apart from that, the thing keeps track of how long it takes the ttx thread
to do its thing, and keeps changing the amount of time between calls to
match how long it takes to do it. so if it takes 10 seconds for the ttx
thread to do its thing, the ttx thread will only get called next time after
10 seconds have passed since it was modified. It's mated like that so
the cpu never gets maxed out.

so keyboard input more keyboard input then the keypresses stop and ten
seconds go by and then the ttx thread will get called, it probably will
take another ten seconds to run, but if it takes less, it will be less time
between the next call to the ttx thread.

there is some minimum time as well, whats the point of burning cpu.

The $specialk token is passed around to keep the modification in sync with
the ttx thread, which means that only runs that are up-to-date with the
master text widget content will get their indices turned into updated tags.

every time the text is modified, the [modified] proc will get called, and
a specialk token will get generated then which makes sure that only the
run that does this last ::only_valid_specialk will get its results used
for the tagging.

}

set after_modified {}
set mtimeout 100  ;# first run timeout, gets changed dynamically later

proc modified {} {
  global master
  global after_modified mtimeout

  set specialk k[clock micro]
  set ::only_valid_specialk $specialk

  after cancel $after_modified
  set after_modified [after $mtimeout [list consider_master $specialk]]

  slot_rerun_was_modified

  $master edit modified 0
}

# all this does is get the master content and send it to the ttx thread.
#
proc consider_master {specialk} {
  global master
  set content [$master get 1.0 end-1c]
  thread::send -async $::TTX [list compute_lol $specialk $content]
}

proc back_to_reality {specialk ztrans ttdiff} {

  global after_modified mtimeout

  # how much time did the run take? even if its not valid use that info
  # to change the modification timeout.
  set n [expr {$ttdiff / 1000}] ; if {$n < 200} {set n 200}
  set mtimeout $n
  ttx_p "back_to_reality => now set to $mtimeout"

  if {$specialk ne $::only_valid_specialk} {
    ttx_p "$specialk isnt valid anymore 9999999999999999999999"
    return
  }

  alter_tags $ztrans

}

proc alter_tags {ztrans} {
  global master
  foreach dd $ztrans {
    dict_set_vars $dd xtag ztx
    $master tag remove $xtag 1.0 end
    if {$ztx eq ""} continue
    $master tag add $xtag {*}$ztx
  }
}

proc ttx_p {x} {

  # the after 10 to free it from deadlock:
  #
  # thread::send to thread::send back and forth between same threads locks it up.
  #
  #
  after 10 [list thread::send $::TTX [list p $x]]

}

















# ompute_lol and compute_lol_aux only run in TTX


set CCC {}

proc compute_lol {specialk content} {
  global CCC
  p "$specialk compute_lol -> after: [after info]"
  after cancel $CCC
  set CCC [after 10 [list compute_lol_aux $specialk $content]]
}

if 0 {

even if its still running, only the last [compute_lol_aux] will get called
after, no matter how many are queued, thanks to the [after 10] 

}


proc compute_lol_aux {specialk content} {
  p "00000000000000000BEGIN $specialk"
  set ttbegin [clock micro]

  set zup {}
  lappend zup [update_proc_right $content]
  lappend zup [update_titles $content]
  lappend zup [update_xiaixa $content]
  ## test of generic: lappend zup [update_generic $content package tag_search]

  reload_komputer $content
  set ztrans {}
  foreach dd $zup {
    dict_set_vars $dd xtag zci
    lappend ztrans [list xtag $xtag ztx [komputer_translate $zci]]
  }

  set ttend [clock micro]
  set ttdiff [expr {$ttend - $ttbegin}]
  p "00000000000000000END $specialk ($ttdiff us)" 

  thread::send $::threadBOSS [list back_to_reality $specialk $ztrans $ttdiff]

}
















# ==title== translate indices with the komputer

text .komputer
proc reload_komputer {c} {
  .komputer del 1.0 end
  .komputer ins end $c
}
proc komputer_translate {z} {
  set zret {}
  foreach ix $z {
    lappend zret [.komputer index "1.0 + $ix c"]
  }
  return $zret
}



# ==title== tagging of text 

proc update_proc_right {c} {

  set xtag tag_proc_right

  set findtext "\nproc "
  set slen [string length $findtext]

  set z {}
  foreach ix [string_find_matches $c $findtext] {

    incr ix $slen

    set i0 $ix

    # read in chars starting from $ix until hitting an invalid char.
    #
    set invalids {"" \n " "}  ;# empty string, newline, space
    set str {}
    #
    while 1 {
      set ch [string index $c $ix]
      if {$ch in $invalids} break
      append str $ch
      incr ix
    }

    set procname $str
    set i1 $ix

    lappend z $i0 $i1
  }

  # content gets searched and returns tag name and list of character indices to tag
  return [list xtag $xtag zci $z]
}


proc update_titles {c} {

  set xtag tag_title

  set findtext [sconcat == title ==]
  set slen [string length $findtext]

  set z {}
  foreach ix [string_find_matches $c $findtext] {

    set i0 [expr {$ix + $slen + 1}]
    set i1 [string first \n $c $i0]

    lappend z $i0 $i1

  }

  return [list xtag $xtag zci $z]
}


proc update_xiaixa {c} {

  set xtag tag_xia

  set findtext [sconcat #### xia ####]
  set slen [string length $findtext]

  set z {}
  foreach ix [string_find_matches $c $findtext] {

    set i0 [expr {$ix + $slen}]
    set i1 [string first \n $c [expr {$i0 + 1}]]
    incr i1 -4

    lappend z $i0 $i1

  }

  return [list xtag $xtag zci $z]
}


proc update_generic {c findtext xtag} {

  set slen [string length $findtext]

  set z {}
  foreach ix [string_find_matches $c $findtext] {

    set i0 $ix
    set i1 [expr {$ix + $slen}]

    lappend z $i0 $i1

  }

  return [list xtag $xtag zci $z]
}



# ==title== more tagging

proc generic_autotag_string {tx findtext xtag} {
  $tx tag remove $xtag 1.0 end
  set c [$tx get 1.0 end-1c]
  foreach ix [string_find_matches $c $findtext] {
    set s0 "1.0 + $ix c"
    set s1 "1.0 + $ix c + [string length $findtext] c"
    $tx tag add $xtag $s0 $s1
  }
}







# ==title== set some defaults


# maybe use the priority font finding stuff from xco
#
set codefamily "Lucida Console"
set codefont {"Lucida Console" 16}
linux {
  # now have lucida console on linux too!
  #set codefont {"DejaVu Sans Mono" 14}
}


proc toggle_fullscreen {w} {
  set t [winfo toplevel $w]
  set x [wm attr $t -fullscreen]
  if {$x} {set x 0} else {set x 1}
  wm attr $t -fullscreen $x
}


# ==title== bindings that apply to all threads and main

bind all <Control-q> exit

bind all <Control-F8> {p [winfo containing [winfo pointerx %W] [winfo pointery %W]]}

bind all <Control-F10> xco_show

bind all <Control-Alt-f> {toggle_fullscreen %W}


windows {
  bind all <Alt-F10> {console show}
}




# ==title== Thread END--------------------------------------!
if {[info exists is_thread]} {
  xco_init
  return
}







# ==title== loaderrrrrrrrrrrrrrrrrr

proc hook_modified {} {
  global master
  bind $master <<Modified>> modified
  modified
}

proc do_the_loadfile {fn} {
  global master universe XFILE

  set XFILE $fn

  # set the title of the toplevel
  #
  set x [file native $fn]
  wm title . "milkjack \[$x\]"

  $master ins end [file_get $fn]

  slots_awake $universe $master

  ## this is replaced by slots
  ##make_nav $universe

  make_new_tx $universe $master

  update

  set zhi {.tcl .tk .tcltk .x}
  if {[file extension $fn] in $zhi} {
  after 10 hook_modified
  }

}


proc build_ui {} {
  global universe master

  set universe .universe
  canvas $universe -highlightthickness 0 -bg #333333
  pack $universe -fill both -expand 1
  init_specials $universe

  set master .master
  text $master
  $master conf -wrap none
  create_master_tags

  maximize_zoom .
  focus -force .
}

proc threads_are_ready {} {
  global universe master argv for_release

  p "[clock micro] threads_are_ready"


  # otherwise focus goes to the last thread initialized!
  focus -force .


  if {$for_release} {

  # ==title== for relase

  windows {
    setup_dnd $universe
  }

  linux {
    if {$argv eq ""} {
      error "on linux the first argument must be the file to edit"
    }
    do_the_loadfile [lindex $argv 0]
  }

  } else {

  
  # ==title== NOT for release

  # edit self

  windows {
    do_the_loadfile milkjack.x
  }

  linux {
    do_the_loadfile milkjack.x
  }


  }

}


# the threads initialize, and in turn call [online_thread]
# which finally calls [threads_are_ready]
#
proc start_threads {} {

  init_ttx_thread
  init_taa_thread

}


set zonline {}
proc online_thread {x} {
  global zonline
  lappend zonline $x

  p "[clock micro] zonline: $zonline"

  if {[lsort $zonline] eq "TAA TTX"} {
    after 10 threads_are_ready
  }
  
}


proc init_ttx_thread {} {

  global TTX
  set TTX [new_thread]
  thread::send $TTX {
    xco_thread_name TTX
    ##xco_show
    after 10 [list thread::send $::threadBOSS [list online_thread TTX]]
  }

}

proc init_taa_thread {} {

  global TAA
  set TAA [new_thread]
  thread::send $TAA {
    xco_thread_name TAA
    ##xco_show
    after 10 [list thread::send $::threadBOSS [list online_thread TAA]]
  }

}



# ==title== run

set for_release 1

package require Thread

set XFILE {}

xco_init

build_ui
update

windows {setup_fonts}
if {$for_release} {make_nfo $universe}

windows {
if {$for_release} {
pack [label .oo -text "welcome home" -font {"Null Pointer;" 48} -bg black -fg red] -fill x
after 2000 [list destroy .oo]
}
}


p "[clock micro] starting threads"

after 10 start_threads


# ==title== more information

# Edit this program with itself by loading milkjack.x into it.

# Control-F10 will show 'my' console, xco, and Alt-F10 will show the tclkit default console on windows.

# enjoy

